import numpy as np
from numba import njit


@njit(cache=True)
def hill_diversity(class_counts: np.ndarray, q: float) -> float:
    '''
    Hill numbers - express actual diversity as opposed e.g. to Gini-Simpson (probability) and Shannon (information)

    exponent at 1 results in undefined because of 1/0 - but limit exists as exp(entropy)
    Ssee "Entropy and diversity" by Lou Jost

    Exponent at 0 = variety - i.e. count of unique species
    Exponent at 1 = unity
    Exponent at 2 = diversity form of simpson index
    '''

    if q < 0:
        raise ValueError('Please select a non-zero value for q.')

    N = class_counts.sum()
    # catch potential division by zero situations
    if N == 0:
        return 0
    # hill number defined in the limit as the exponential of information entropy
    if q == 1:
        H = 0
        for a in class_counts:
            if a:  # if not 0
                p = a / N  # the probability of this class
                H += p * np.log(p)  # sum entropy
        return np.exp(-H)  # return exponent of entropy
    # otherwise use the usual form of Hill numbers
    else:
        D = 0
        for a in class_counts:
            if a:
                p = a / N  # the probability of this class
                D += p ** q  # sum
        return D ** (1 / (1 - q))  # return as equivalent species


@njit(cache=True)
def hill_diversity_branch_distance_wt(class_counts: np.ndarray,
                                      class_distances: np.array,
                                      q: float,
                                      beta: float) -> float:
    '''
    Based on unified framework for species diversity in Chao, Chiu, Jost 2014
    See table on page 308 and surrounding text

    In this case the presumption is that you supply branch weights in the form of negative exponential distance weights
    i.e. pedestrian walking distance decay which weights more distant locations more weakly than nearer locations
    This means that the walking distance to a landuse impacts how strongly it contributes to diversity

    The weighting is based on the nearest of each landuse
    This is debatably most relevant to q=0
    '''

    if len(class_counts) != len(class_distances):
        raise ValueError('Mismatching number of unique class counts and respective class distances.')

    if beta > 0:
        raise ValueError('Please provide the beta with the leading negative.')

    if q < 0:
        raise ValueError('Please select a non-zero value for q.')

    # catch potential division by zero situations
    N = class_counts.sum()
    if N == 0:
        return 0

    # find T
    T = 0
    for i in range(len(class_counts)):
        if class_counts[i]:
            a = class_counts[i] / N
            wt = np.exp(class_distances[i] * beta)
            T += wt * a

    # hill number defined in the limit as the exponential of information entropy
    if q == 1:
        PD_lim = 0  # using the same variable name as non limit version causes errors for parallel
        # get branch lengths and class abundances
        for i in range(len(class_counts)):
            if class_counts[i]:
                a = class_counts[i] / N
                wt = np.exp(class_distances[i] * beta)
                PD_lim += wt * a / T * np.log(a / T)  # sum entropy
        # return exponent of entropy
        PD_lim = np.exp(-PD_lim)
        return PD_lim  # / T
    # otherwise use the usual form of Hill numbers
    else:
        PD = 0
        # get branch lengths and class abundances
        for i in range(len(class_counts)):
            if class_counts[i]:
                a = class_counts[i] / N
                wt = np.exp(class_distances[i] * beta)
                PD += wt * (a / T) ** q  # sum
        # once summed, apply q
        PD = PD ** (1 / (1 - q))
        return PD  # / T


@njit(cache=True)
def hill_diversity_pairwise_distance_wt(class_counts: np.ndarray,
                                        class_distances: np.ndarray,
                                        q: float,
                                        beta: float) -> float:
    '''
    This is the distances version - see below for disparity matrix version

    Based on unified framework for species diversity in Chao, Chiu, Jost 2014
    See table on page 308 and surrounding text

    In this case the presumption is that you supply branch weights in the form of negative exponential distance weights
    i.e. pedestrian walking distance decay which weights more distant locations more weakly than nearer locations
    This means that the walking distance to a landuse impacts how strongly it contributes to diversity

    Functional diversity takes the pairwise form, thus distances are based on pairwise i to j distances via the node k
    Remember these are already distilled species counts - so it is OK to use closest distance to each species

    This is different to the non-pairwise form of the phylogenetic version which simply takes singular distance k to i
    '''

    if len(class_counts) != len(class_distances):
        raise ValueError('Mismatching number of unique class counts and respective class distances.')

    if beta > 0:
        raise ValueError('Please provide the beta with the leading negative.')

    if q < 0:
        raise ValueError('Please select a non-zero value for q.')

    # catch potential division by zero situations
    N = class_counts.sum()
    if N == 0:
        return 0

    # calculate Q
    Q = 0
    for i in range(len(class_counts)):
        if class_counts[i]:
            a_i = class_counts[i] / N
            for j in range(len(class_counts)):
                # only need to examine the pair if j < i, otherwise double-counting
                if j > i:
                    break
                if class_counts[j]:
                    a_j = class_counts[j] / N
                    wt = np.exp((class_distances[i] + class_distances[j]) * beta)
                    # pairwise distances
                    Q += wt * a_i * a_j

    # pairwise disparities weights can sometimes give rise to Q = 0... causing division by zero etc.
    if Q == 0:
        return 0

    # if in the limit, use exponential
    if q == 1:
        FD_lim = 0  # using the same variable name as non limit version causes errors for parallel
        for i in range(len(class_counts)):
            if class_counts[i]:
                a_i = class_counts[i] / N
                for j in range(len(class_counts)):
                    # only need to examine the pair if j < i, otherwise double-counting
                    if j > i:
                        break
                    if class_counts[j]:
                        a_j = class_counts[j] / N
                        # pairwise distances
                        wt = np.exp((class_distances[i] + class_distances[j]) * beta)
                        FD_lim += wt * a_i * a_j / Q * np.log(a_i * a_j / Q)  # sum
        # once summed
        FD_lim = np.exp(-FD_lim)
        return FD_lim ** (1 / 2)  # (FD_lim / Q) ** (1 / 2)
    # otherwise conventional form
    else:
        FD = 0
        for i in range(len(class_counts)):
            if class_counts[i]:
                a_i = class_counts[i] / N
                for j in range(len(class_counts)):
                    # only need to examine the pair if j < i, otherwise double-counting
                    if j > i:
                        break
                    if class_counts[j]:
                        a_j = class_counts[j] / N
                        # pairwise distances
                        wt = np.exp((class_distances[i] + class_distances[j]) * beta)
                        FD += wt * (a_i * a_j / Q) ** q  # sum
        FD = FD ** (1 / (1 - q))
        return FD ** (1 / 2)  # (FD / Q) ** (1 / 2)


@njit(cache=True)
def hill_diversity_pairwise_matrix_wt(class_counts: np.ndarray, wt_matrix: np.ndarray, q: float) -> float:
    '''
    This is the matrix version - requires a precomputed (e.g. disparity) matrix for all classes.

    See above for distance version.
    '''

    if len(class_counts) != len(wt_matrix):
        raise ValueError('Mismatching number of unique class counts and dimensionality of class weights matrix.')

    if not wt_matrix.ndim == 2 or wt_matrix.shape[0] != wt_matrix.shape[1]:
        raise ValueError('Weights matrix must be an NxN pairwise matrix of disparity weights.')

    if q < 0:
        raise ValueError('Please select a non-zero value for q.')

    # catch potential division by zero situations
    N = class_counts.sum()
    if N == 0:
        return 0

    # calculate Q
    Q = 0
    for i in range(len(class_counts)):
        if class_counts[i]:
            a_i = class_counts[i] / N
            for j in range(len(class_counts)):
                # only need to examine the pair if j < i, otherwise double-counting
                if j > i:
                    break
                if class_counts[j]:
                    a_j = class_counts[j] / N
                    wt = wt_matrix[i][j]
                    # pairwise distances
                    Q += wt * a_i * a_j

    # pairwise disparities weights can sometimes give rise to Q = 0... causing division by zero etc.
    if Q == 0:
        return 0

    # if in the limit, use exponential
    if q == 1:
        FD_lim = 0  # using the same variable name as non limit version causes errors for parallel
        for i in range(len(class_counts)):
            if class_counts[i]:
                a_i = class_counts[i] / N
                for j in range(len(class_counts)):
                    # only need to examine the pair if j < i, otherwise double-counting
                    if j > i:
                        break
                    if class_counts[j]:
                        a_j = class_counts[j] / N
                        # pairwise distances
                        wt = wt_matrix[i][j]
                        FD_lim += wt * a_i * a_j / Q * np.log(a_i * a_j / Q)  # sum
        # once summed
        FD_lim = np.exp(-FD_lim)
        return FD_lim ** (1 / 2)  # (FD_lim / Q) ** (1 / 2)
    # otherwise conventional form
    else:
        FD = 0
        for i in range(len(class_counts)):
            if class_counts[i]:
                a_i = class_counts[i] / N
                for j in range(len(class_counts)):
                    # only need to examine the pair if j < i, otherwise double-counting
                    if j > i:
                        break
                    if class_counts[j]:
                        a_j = class_counts[j] / N
                        # pairwise distances
                        wt = wt_matrix[i][j]
                        FD += wt * (a_i * a_j / Q) ** q  # sum
        FD = FD ** (1 / (1 - q))
        return FD ** (1 / 2)  # (FD / Q) ** (1 / 2)


@njit(cache=True)
def gini_simpson_diversity(class_counts: np.ndarray) -> float:
    '''
    Gini-Simpson
    Gini transformed to 1 − λ
    Probability that two individuals picked at random do not represent the same species (Tuomisto)

    Ordinarily:
    D = 1 - sum(p**2) where p = Xi/N

    Bias corrected:
    D = 1 - sum(Xi/N * (Xi-1/N-1))
    '''
    N = class_counts.sum()
    G = 0
    # catch potential division by zero situations
    if N < 2:
        return G
    # compute bias corrected gini-simpson
    for c in class_counts:
        if c:
            G += c / N * (c - 1) / (N - 1)
    return 1 - G


@njit(cache=True)
def shannon_diversity(class_counts: np.ndarray) -> float:
    '''
    Entropy
    p = Xi/N
    S = -sum(p * log(p))
    Uncertainty of the species identity of an individual picked at random (Tuomisto)
    '''
    N = class_counts.sum()
    H = 0
    # catch potential division by zero situations
    if N == 0:
        return H
    # compute
    for a in class_counts:
        if a:
            p = a / N  # the probability of this class
            H += p * np.log(p)  # sum entropy
    return -H  # remember negative


@njit(cache=True)
def raos_quadratic_diversity(class_counts: np.ndarray,
                             wt_matrix: np.array,
                             alpha: float = 1,
                             beta: float = 1) -> float:
    '''
    Rao's quadratic - bias corrected and based on disparity

    Sum of weighted pairwise products

    Note that Stirling's diversity is a rediscovery of Rao's quadratic diversity
    Though adds alpha and beta exponents to tweak weights of disparity dij and pi * pj, respectively
    This is a hybrid of the two, i.e. including alpha and beta options and adjusted for bias
    Rd = sum(dij * Xi/N * (Xj/N-1))

    Behaviour is controlled using alpha and beta exponents
    0 and 0 reduces to variety (effectively a count of unique types)
    0 and 1 reduces to balance (half-gini - pure balance, no weights)
    1 and 0 reduces to disparity (effectively a weighted count)
    1 and 1 is base stirling diversity / raos quadratic
    '''

    if len(class_counts) != len(wt_matrix):
        raise ValueError('Mismatching number of unique class counts and respective class taxonomy tiers.')

    if not wt_matrix.ndim == 2 or wt_matrix.shape[0] != wt_matrix.shape[1]:
        raise ValueError('Weights matrix must be an NxN pairwise matrix of disparity weights.')

    # catch potential division by zero situations
    N = class_counts.sum()
    if N < 2:
        return 0

    R = 0  # variable for additive calculations of distance * p1 * p2
    for i in range(len(class_counts)):
        # parallelise only inner loop
        for j in range(len(class_counts)):
            # only need to examine the pair if j <= i, otherwise double-counting
            if j > i:
                break
            if class_counts[i] and class_counts[j]:
                p_i = class_counts[i] / N  # place here to catch division by zero for single element
                p_j = class_counts[j] / (N - 1)  # bias adjusted
                if p_i and p_j:  # if the probabilities aren't 0
                    # calculate 3rd level disparity
                    wt = wt_matrix[i][j]
                    R += wt ** alpha * (p_i * p_j) ** beta
    return R



'''
https://docs.python.org/3.9/distutils/examples.html
https://docs.python.org/3/distutils/setupscript.html
https://scikit-build.readthedocs.io/en/latest/usage.html

# manual deployment
pip install --upgrade setuptools wheel
python setup.py sdist bdist_wheel
pip install --upgrade twine
TESTING REPO: twine upload --repository-url https://test.pypi.org/legacy/ dist/*
DOWNLOADING FROM TEST REPO: pip install --extra-index-url https://test.pypi.org/simple/ cityseer==1.0.0a1
OTHERWISE: twine upload dist/*
'''

from setuptools import setup

setup(
    name='cityseer',
    version='1.0.0alpha3',
    packages=['cityseer', 'cityseer.algos', 'cityseer.metrics', 'cityseer.tools'],
    description='Computational tools for urban analysis',
    url='https://github.com/cityseer/cityseer',
    project_urls={
        "Bug Tracker": "https://github.com/cityseer/cityseer/issues",
        "Documentation": "https://cityseer.github.io/cityseer/",
        "Source Code": "https://github.com/cityseer/cityseer",
    },
    author='Gareth Simons',
    author_email='gareth@cityseer.io',
    license='Apache 2.0 + "Commons Clause" License Condition v1.0',
    install_requires=[
        'numpy',
        'numba',
        'utm',
        'shapely>=1.7.0',
        'networkx>=2.4',
        'tqdm',
        'matplotlib',
        'sklearn'
    ]
)



"""
Convenience methods for plotting graphs within the cityseer API context. This module is predominately used for basic
plots or visual verification of behaviour in code tests. Custom behaviour can be achieved by directly manipulating the
underlying [`NetworkX`](https://networkx.github.io) and [`matplotlib`](https://matplotlib.org) figures.
"""
import logging
from typing import Union

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from matplotlib import colors
from matplotlib.collections import LineCollection
from shapely import geometry
from sklearn.preprocessing import LabelEncoder

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

primary = '#0091ea'
accent = '#64c1ff'
info = '#0064b7'
secondary = '#d32f2f'
warning = '#9a0007'
error = '#ffffff'
success = '#2e7d32'
background = '#2e2e2e'


def plot_nX_primal_or_dual(primal_graph: nx.MultiGraph = None,
                           dual_graph: nx.MultiGraph = None,
                           path: str = None,
                           labels: bool = False,
                           primal_node_colour: Union[str, tuple, list] = None,
                           primal_edge_colour: str = None,
                           dual_node_colour: Union[str, tuple, list] = None,
                           dual_edge_colour: str = None,
                           primal_edge_width: Union[int, float] = None,
                           dual_edge_width: Union[int, float] = None,
                           plot_geoms: bool = True,
                           x_lim: Union[tuple, list] = None,
                           y_lim: Union[tuple, list] = None,
                           **figure_kwargs):
    """
    Parameters
    ----------
    primal_graph
        An optional `NetworkX` MultiGraph to plot in the primal representation. Defaults to None.
    dual_graph
        An optional `NetworkX` MultiGraph to plot in the dual representation. Defaults to None.
    path
        An optional filepath: if provided, the image will be saved to the path instead of being displayed. Defaults to
        None.
    labels
        Whether to display node labels. Defaults to False.
    primal_node_colour
        Primal node colour or colours. When passing an iterable of colours, the number of colours should match the order
        and number of nodes in the MultiGraph. The colours are passed to the underlying [`draw_networkx`](https://networkx.github.io/documentation/networkx-1.10/reference/generated/networkx.drawing.nx_pylab.draw_networkx.html#draw-networkx)
        method and should be formatted accordingly. Defaults to None.
    primal_edge_colour
        Primal edge colour as a `matplotlib` compatible colour string. Defaults to None.
    dual_node_colour
        Dual node colour or colours. When passing a list of colours, the number of colours should match the order and
        number of nodes in the MultiGraph. The colours are passed to the underlying [`draw_networkx`](https://networkx.github.io/documentation/networkx-1.10/reference/generated/networkx.drawing.nx_pylab.draw_networkx.html#draw-networkx)
        method and should be formatted accordingly. Defaults to None.
    dual_edge_colour
        Dual edge colour as a `matplotlib` compatible colour string. Defaults to None.
    primal_edge_width
        Linewidths for the primal edge. Defaults to None.
    dual_edge_width
        Linewidths for the dual edge. Defaults to None.
    plot_geoms
        Whether to plot the edge geometries. If set to `False`, straight lines will be drawn from node-to-node to
        represent edges. Defaults to True.
    x_lim
        A tuple or list with the minimum and maxium `x` extents to be plotted.
        Defaults to None.
    y_lim
        A tuple or list with the minimum and maxium `y` extents to be plotted.
        Defaults to None.
    figure_kwargs
        `kwargs` which will be passed to the `matplotlib` figure parameters. If
        provided, these will override the default figure size or dpi parameters.

    Notes
    -----
    Plot either or both primal and dual representations of a `networkX MultiGraph`. Only call this function directly if
    explicitly printing both primal and dual graphs. Otherwise, use the simplified [`plot_nX`](plot#plot-nx) method
    instead.

    ```py
    from cityseer.tools import mock, graphs, plot
    G = mock.mock_graph()
    G_simple = graphs.nX_simple_geoms(G)
    G_dual = graphs.nX_to_dual(G_simple)
    plot.plot_nX_primal_or_dual(G_simple,
                                G_dual,
                                plot_geoms=False)
    ```
    ![Example primal and dual graph plot.](../.vitepress/plots/images/graph_dual.png)
    _A dual graph in blue overlaid on the source primal graph in red._
    """
    # cleanup old plots
    plt.ioff()
    plt.close('all')
    plt.cla()
    plt.clf()
    # create new plot
    fig, ax = plt.subplots(1, 1, **figure_kwargs)
    # setup params
    alpha = 0.75
    node_size = 30
    if labels:
        alpha = 0.95
        node_size = 70

    # setup a function that can be used for either the primal or dual graph
    def plot_graph(_graph, _is_primal, _node_colour, _node_shape, _edge_colour, _edge_style, _edge_width):
        if not len(_graph.nodes()):
            raise ValueError('Graph contains no nodes to plot.')
        iterables = (list, tuple, np.ndarray)
        if _node_colour is not None:
            if isinstance(_node_colour, iterables) and len(_node_colour) != len(_graph):
                raise ValueError('A list, tuple, or array of colours should match the number of nodes in the graph.')
        else:
            if _is_primal:
                _node_colour = secondary
            else:
                _node_colour = info
        if _edge_colour is not None:
            if not isinstance(_edge_colour, str):
                raise ValueError('Edge colours should be a string representing a single colour.')
        else:
            if _is_primal:
                _edge_colour = 'w'
            else:
                _edge_colour = accent
        # edge width
        if _edge_width is not None:
            if not isinstance(_edge_width, (int, float)):
                raise ValueError('Edge widths should be an int or float.')
        else:
            if _is_primal:
                _edge_width = 1.5
            else:
                _edge_width = 0.5
        pos = {}
        node_list = []
        # setup a node colour list if nodes are individually coloured, this is for filtering by extents
        colour_list = []
        for n_idx, (n, d) in enumerate(_graph.nodes(data=True)):
            x = d['x']
            y = d['y']
            # add to the pos dictionary regardless (otherwise nx.draw throws an error)
            pos[n] = (d['x'], d['y'])
            # filter out nodes not within the extnets
            if x_lim and (x < x_lim[0] or x > x_lim[1]):
                continue
            if y_lim and (y < y_lim[0] or y > y_lim[1]):
                continue
            # add to the node list
            node_list.append(n)
            # and add to the node colour list if node colours are a list or tuple
            if isinstance(_node_colour, iterables):
                colour_list.append(_node_colour[n_idx])
        if not len(node_list):
            raise ValueError('All nodes have been filtered out by the x_lim / y_lim parameter: check your extents')
        # update the node colours to the filtered list of colours if necessary
        if isinstance(_node_colour, iterables):
            _node_colour = colour_list
        # plot edges manually for geoms
        edge_list = []
        edge_geoms = []
        for s, e, data in _graph.edges(data=True):
            # filter out if start and end nodes are not in the active node list
            if s not in node_list and e not in node_list:
                continue
            if plot_geoms:
                try:
                    x_arr, y_arr = data['geom'].coords.xy
                except KeyError:
                    raise KeyError(f"Can't plot geoms because a 'geom' key can't be found for edge {s} to {e}. "
                                   f"Use the nX_simple_geoms() method if you need to create geoms for a graph.")
                edge_geoms.append(tuple(zip(x_arr, y_arr)))
            else:
                edge_list.append((s, e))
        # plot geoms manually if required
        if plot_geoms:
            lines = LineCollection(edge_geoms, colors=_edge_colour, linewidths=_edge_width, linestyles=_edge_style)
            ax.add_collection(lines)
        # go ahead and plot: note that edge_list will be empty if plotting geoms
        nx.draw(_graph,
                pos=pos,
                ax=ax,
                with_labels=labels,
                font_size=5,
                font_color='w',
                font_weight='bold',
                nodelist=node_list,
                node_color=_node_colour,
                node_size=node_size,
                node_shape=_node_shape,
                edgelist=edge_list,
                edge_color=_edge_colour,
                style=_edge_style,
                width=_edge_width,
                alpha=alpha)

    if primal_graph is not None:
        plot_graph(primal_graph, True, primal_node_colour, 'o', primal_edge_colour, 'solid', primal_edge_width)
    if dual_graph is not None:
        plot_graph(dual_graph, False, dual_node_colour, 'd', dual_edge_colour, 'dashed', dual_edge_width)
    if x_lim:
        plt.xlim(*x_lim)
    if y_lim:
        plt.ylim(*y_lim)
    if path:
        plt.savefig(path, facecolor=background)
    else:
        plt.gcf().set_facecolor(background)
        plt.show()


def plot_nX(networkX_graph: nx.MultiGraph,
            path: str = None,
            labels: bool = False,
            node_colour: Union[str, tuple, list] = None,
            edge_colour: Union[str, tuple, list] = None,
            edge_width: Union[int, float] = None,
            plot_geoms: bool = False,
            x_lim: Union[tuple, list] = None,
            y_lim: Union[tuple, list] = None,
            **figure_kwargs):
    """
    Plot a `networkX` MultiGraph.

    Parameters
    ----------
    networkX_graph
        A `NetworkX` MultiGraph.
    path
        An optional filepath: if provided, the image will be saved to the path instead of being displayed. Defaults to
        None.
    labels
        Whether to display node labels. Defaults to False.
    node_colour
        Node colour or colours. When passing an iterable of colours, the number of colours should match the order and
        number of nodes in the MultiGraph. The colours are passed to the underlying [`draw_networkx`](https://networkx.github.io/documentation/networkx-1.10/reference/generated/networkx.drawing.nx_pylab.draw_networkx.html#draw-networkx)
        method and should be formatted accordingly. Defaults to None.
    edge_colour
        Edges colour as a `matplotlib` compatible colour string. Defaults to None.
    edge_width
        Linewidths for edges. Defaults to None.
    plot_geoms
        Whether to plot the edge geometries. If set to `False`, straight lines will be drawn from node-to-node to
        represent edges. Defaults to True.
    x_lim
        A tuple or list with the minimum and maxium `x` extents to be plotted. Defaults to None.
    y_lim
        A tuple or list with the minimum and maxium `y` extents to be plotted. Defaults to None.
    figure_kwargs
        `kwargs` which will be passed to the `matplotlib` figure parameters. If provided, these will override the
        default figure size or dpi parameters.

    Notes
    -----
    ```py
    from cityseer.tools import mock, graphs, plot
    from cityseer.metrics import networks
    from matplotlib import colors
    # generate a MultiGraph and compute gravity
    G = mock.mock_graph()
    G = graphs.nX_simple_geoms(G)
    G = graphs.nX_decompose(G, 50)
    N = networks.NetworkLayerFromNX(G, distances=[800])
    N.node_centrality(measures=['node_beta'])
    G_after = N.to_networkX()
    # let's extract and normalise the values
    vals = []
    for node, data in G_after.nodes(data=True):
        vals.append(data['metrics']['centrality']['node_beta'][800])
    # let's create a custom colourmap using matplotlib
    cmap = colors.LinearSegmentedColormap.from_list('cityseer',
                                                    [(100/255, 193/255, 255/255, 255/255),
                                                    (211/255, 47/255, 47/255, 1/255)])
    # normalise the values
    vals = colors.Normalize()(vals)
    # cast against the colour map
    cols = cmap(vals)
    # plot
    plot.plot_nX(G_after, node_colour=cols)
    ```

    ![Example Colour Plot.](../.vitepress/plots/images/graph_colour.png)
    _Colour plot of 800m gravity index centrality on a 50m decomposed graph._
    """
    return plot_nX_primal_or_dual(primal_graph=networkX_graph,
                                  path=path,
                                  labels=labels,
                                  primal_node_colour=node_colour,
                                  primal_edge_colour=edge_colour,
                                  primal_edge_width=edge_width,
                                  plot_geoms=plot_geoms,
                                  x_lim=x_lim,
                                  y_lim=y_lim,
                                  **figure_kwargs)


def plot_assignment(Network_Layer,
                    Data_Layer,
                    path: str = None,
                    node_colour: Union[list, tuple, np.ndarray] = None,
                    node_labels: bool = False,
                    data_labels: Union[list, tuple, np.ndarray] = None,
                    **figure_kwargs):
    """
    Plot a `NetworkLayer` and `DataLayer` for the purpose of visualising assignment of data points to respective nodes.

    Parameters
    ----------
    Network_Layer
        A [`NetworkLayer`](/metrics/networks#class-networklayer).
    Data_Layer
        A [`DataLayer`](/metrics/layers#class-datalayer).
    path
        An optional filepath: if provided, the image will be saved to the path instead of being displayed. Defaults to
        None.
    node_colour
        Node colour or colours. When passing a list of colours, the number of colours should match the order and number
        of nodes in the MultiGraph. The colours are passed to the underlying [`draw_networkx`](https://networkx.github.io/documentation/networkx-1.10/reference/generated/networkx.drawing.nx_pylab.draw_networkx.html#draw-networkx)
        method and should be formatted accordingly. Defaults to None.
    node_labels
        Whether to plot the node labels. Defaults to False.
    data_labels
        An optional iterable of categorical data labels which will be mapped to colours. The number of labels should
        match the number of data points in `DataLayer`. Defaults to None.
    figure_kwargs
        `kwargs` which will be passed to the `matplotlib` figure parameters. If provided, these will override the
        default figure size or dpi parameters.

    Notes
    -----
    ![Example assignment plot.](../.vitepress/plots/images/assignment_plot.png)
    _An assignment plot to a 50m decomposed graph, with the data points coloured by categorical labels._
    """
    plt.figure(**figure_kwargs)

    # extract NetworkX
    Graph = Network_Layer.to_networkX()

    if node_colour is not None:
        if not (len(node_colour) == 1 or len(node_colour) == len(Graph)):
            raise ValueError('Node colours should either be a single colour or a list or tuple of colours matching '
                             'the number of nodes in the graph.')
        node_colour = node_colour
    else:
        node_colour = secondary

    # do a simple plot - don't provide path
    pos = {}
    for n, d in Graph.nodes(data=True):
        pos[n] = (d['x'], d['y'])
    nx.draw(Graph, pos,
            with_labels=node_labels,
            font_size=5,
            font_color='w',
            font_weight='bold',
            node_color=node_colour,
            node_size=30,
            node_shape='o',
            edge_color='w',
            width=1,
            alpha=0.75)

    if data_labels is None:
        data_colour = info
        data_cmap = None
    else:
        # generate categorical colormap
        # use sklearn's label encoder
        le = LabelEncoder()
        le.fit(data_labels)
        # map the int encodings to the respective classes
        classes_int = le.transform(data_labels)
        data_colour = colors.Normalize()(classes_int)
        data_cmap = 'Dark2'  # Set1

    # overlay data map
    plt.scatter(x=Data_Layer._data[:, 0],
                y=Data_Layer._data[:, 1],
                c=data_colour,
                cmap=data_cmap,
                s=30,
                edgecolors='white',
                lw=0.5,
                alpha=0.95)

    # draw assignment
    for i, (x, y, nearest_netw_idx, next_n_netw_idx) in \
            enumerate(zip(Data_Layer._data[:, 0],
                          Data_Layer._data[:, 1],
                          Data_Layer._data[:, 2],
                          Data_Layer._data[:, 3])):

        # if the data points have been assigned network indices
        if not np.isnan(nearest_netw_idx):
            # plot lines to parents for easier viz
            p_x = Network_Layer._node_data[int(nearest_netw_idx)][0]
            p_y = Network_Layer._node_data[int(nearest_netw_idx)][1]
            plt.plot([p_x, x], [p_y, y], c='#64c1ff', lw=0.5, ls='--')

        if not np.isnan(next_n_netw_idx):
            p_x = Network_Layer._node_data[int(next_n_netw_idx)][0]
            p_y = Network_Layer._node_data[int(next_n_netw_idx)][1]
            plt.plot([p_x, x], [p_y, y], c='#888888', lw=0.5, ls='--')

    if path:
        plt.savefig(path, facecolor=background, dpi=150)
    else:
        plt.gcf().set_facecolor(background)
        plt.show()


def plot_graph_maps(node_data: np.ndarray,
                    edge_data: np.ndarray,
                    data_map: np.ndarray = None,
                    poly: geometry.Polygon = None):
    """
    Plot a graph from raw `cityseer` data structures. Note that this function is subject to frequent revision pending
    short-term development requirements. It is used mainly to visually confirm the correct behaviour of particular
    algorithms during the software development cycle.

    Parameters
    ----------
    node_data
        `cityseer` node map.
    edge_data
        `cityseer` edge map.
    data_map
        An optional data map. Defaults to None.
    poly
        An optional polygon. Defaults to None.
    """
    # the edges are bi-directional - therefore duplicated per directional from-to edge
    # use two axes to check each copy of edges
    fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(6, 10))
    # set extents
    for ax in (ax1, ax2):
        ax.set_xlim(node_data[:, 0].min() - 100, node_data[:, 0].max() + 100)
        ax.set_ylim(node_data[:, 1].min() - 100, node_data[:, 1].max() + 100)
    if poly:
        x = [x for x in poly.exterior.coords.xy[0]]
        y = [y for y in poly.exterior.coords.xy[1]]
        ax1.plot(x, y)
        ax2.plot(x, y)
    # plot nodes
    cols = []
    for n in node_data[:, 2]:
        if bool(n):
            cols.append(secondary)
        else:
            cols.append(accent)
    ax1.scatter(node_data[:, 0], node_data[:, 1], s=30, c=primary, edgecolor=cols, lw=0.5)
    ax2.scatter(node_data[:, 0], node_data[:, 1], s=30, c=primary, edgecolor=cols, lw=0.5)
    # plot edges
    processed_edges = set()
    for edge_idx, edge_data in enumerate(edge_data):
        s, e = edge_data[:2]
        se_key = '-'.join(sorted([str(s), str(e)]))
        # bool indicating whether second copy in opposite direction
        dupe = se_key in processed_edges
        processed_edges.add(se_key)
        # get the start and end coords - this is to check that still within src edge - neighbour range
        s_x, s_y = node_data[int(s), :2]
        e_x, e_y = node_data[int(e), :2]
        if not dupe:
            ax1.plot([s_x, e_x], [s_y, e_y], c=accent, linewidth=1)
        else:
            ax2.plot([s_x, e_x], [s_y, e_y], c=accent, linewidth=1)
    for node_idx, n_data in enumerate(node_data):
        ax2.annotate(node_idx, xy=n_data[:2], size=5)
    '''
    DATA MAP:
    0 - x
    1 - y
    2 - assigned network index - nearest
    3 - assigned network index - next-nearest
    '''
    if data_map is not None:
        # plot parents on ax1
        ax1.scatter(x=data_map[:, 0],
                    y=data_map[:, 1],
                    color=secondary,
                    edgecolor=warning,
                    alpha=0.9,
                    lw=0.5)
        ax2.scatter(x=data_map[:, 0],
                    y=data_map[:, 1],
                    color=secondary,
                    edgecolor=warning,
                    alpha=0.9,
                    lw=0.5)
        for i, assignment_data in enumerate(data_map):
            x, y, nearest_netw_idx, next_n_netw_idx = assignment_data
            ax2.annotate(str(int(i)), xy=(x, y), size=8, color='red')
            # if the data points have been assigned network indices
            if not np.isnan(nearest_netw_idx):
                # plot lines to parents for easier viz
                p_x = node_data[int(nearest_netw_idx)][0]
                p_y = node_data[int(nearest_netw_idx)][1]
                ax1.plot([p_x, x], [p_y, y], c=warning, lw=0.75, ls='-')
            if not np.isnan(next_n_netw_idx):
                p_x = node_data[int(next_n_netw_idx)][0]
                p_y = node_data[int(next_n_netw_idx)][1]
                ax1.plot([p_x, x], [p_y, y], c=info, lw=0.75, ls='--')
    plt.gcf().set_facecolor(background)
    plt.show()
